@@ -5,7 +5,7 @@
 ;; Author: Feng Shu <tumashu@163.com>
 ;; Maintainer: Feng Shu <tumashu@163.com>
 ;; URL: https://github.com/tumashu/vertico-posframe
-;; Version: 0.5.0
+;; Version: 0.6.0
 ;; Keywords: abbrev, convenience, matching, vertico
 ;; Package-Requires: ((emacs "26.0") (posframe "1.1.4") (vertico "0.13.0"))
 
@@ -36,6 +36,22 @@
 ;; NOTE: vertico-posframe requires Emacs 26 and do not support mouse
 ;; click.
 
+;; ** How to let vertico-posframe work well with vertico-multiform.
+;; #+BEGIN_EXAMPLE
+;; (setq vertico-multiform-commands
+;;       '((consult-line
+;;          posframe
+;;          (vertico-posframe-poshandler . posframe-poshandler-frame-top-center)
+;;          (vertico-posframe-border-width . 10)
+;;          ;; NOTE: This is useful when emacs is used in both in X and
+;;          ;; terminal, for posframe do not work well in terminal, so
+;;          ;; vertico-buffer-mode will be used as fallback at the
+;;          ;; moment.
+;;          (vertico-posframe-fallback-mode . vertico-buffer-mode))
+;;         (t posframe)))
+;; (vertico-multiform-mode 1)
+;; #+END_EXAMPLE
+
 ;; ** How to enable vertico-posframe
 ;; #+BEGIN_EXAMPLE
 ;; (require 'vertico-posframe)
@@ -61,7 +77,11 @@
 
 (defgroup vertico-posframe nil
   "Using posframe to show vertico."
-  :group 'vertico-posframe)
+  :group 'vertico)
+
+(defcustom vertico-posframe-fallback-mode #'ignore
+  "When posframe is not workable, this mode will be used as fallback."
+  :type 'function)
 
 (defcustom vertico-posframe-font nil
   "The font used by vertico-posframe.
@@ -84,6 +104,10 @@ When nil, Using current frame's font as fallback."
   "The min height of vertico-posframe."
   :type 'number)
 
+(defcustom vertico-posframe-truncate-lines t
+  "Non-nil means truncate lines in vertico-posframe."
+  :type 'boolean)
+
 (defcustom vertico-posframe-poshandler #'posframe-poshandler-frame-center
   "The posframe poshandler used by vertico-posframe."
   :type 'function)
@@ -127,7 +151,27 @@ minibuffer will not be hided by minibuffer-cover."
 
 (defface vertico-posframe-border
   '((t (:inherit default :background "gray50")))
-  "Face used by the vertico-posframe's border."
+  "Face used by the vertico-posframe's border when minibuffer-depth = 1."
+  :group 'vertico-posframe)
+
+(defface vertico-posframe-border-2
+  '((t (:inherit default :background "red")))
+  "Face used by the vertico-posframe's border when minibuffer-depth = 2."
+  :group 'vertico-posframe)
+
+(defface vertico-posframe-border-3
+  '((t (:inherit default :background "green")))
+  "Face used by the vertico-posframe's border when minibuffer-depth = 3."
+  :group 'vertico-posframe)
+
+(defface vertico-posframe-border-4
+  '((t (:inherit default :background "blue")))
+  "Face used by the vertico-posframe's border when minibuffer-depth = 4."
+  :group 'vertico-posframe)
+
+(defface vertico-posframe-border-fallback
+  '((t (:inherit default :background "yellow")))
+  "Face used by the vertico-posframe's border when find no face."
   :group 'vertico-posframe)
 
 (defvar vertico-posframe--buffer nil)
@@ -160,55 +204,57 @@ Optional argument FRAME ."
   "Hidehandler used by vertico-posframe."
   (not (minibufferp)))
 
-(defun vertico-posframe-get-size ()
+(defun vertico-posframe-get-size (buffer)
   "The default functon used by `vertico-posframe-size-function'."
   (list
-   :height vertico-posframe-height
-   :width vertico-posframe-width
-   :min-height (or vertico-posframe-min-height
+   :height (buffer-local-value 'vertico-posframe-height buffer)
+   :width (buffer-local-value 'vertico-posframe-width buffer)
+   :min-height (or (buffer-local-value 'vertico-posframe-min-height buffer)
                    (let ((height (+ vertico-count 1)))
-                     (min height (or vertico-posframe-height height))))
-   :min-width (or vertico-posframe-min-width
+                     (min height (or (buffer-local-value 'vertico-posframe-height buffer) height))))
+   :min-width (or (buffer-local-value 'vertico-posframe-min-width buffer)
                   (let ((width (round (* (frame-width) 0.62))))
-                    (min width (or vertico-posframe-width width))))))
+                    (min width (or (buffer-local-value 'vertico-posframe-width buffer) width))))))
 
 (defun vertico-posframe--display (_lines)
-  "Display LINES in posframe."
-  (let* ((show-minibuffer-p (vertico-posframe--show-minibuffer-p))
-         (minibuffer-window (active-minibuffer-window))
-         (point (point)))
-    (setq vertico-posframe--buffer (current-buffer))
-    (window-resize minibuffer-window
-                   (- (window-pixel-height minibuffer-window))
-                   nil nil 'pixelwise)
-    (set-window-vscroll minibuffer-window 100)
-    (when show-minibuffer-p
-      (set-window-vscroll minibuffer-window 0))
-    (with-selected-window (vertico-posframe-last-window)
-      (vertico-posframe--show vertico-posframe--buffer point))))
+  "Display _LINES in posframe."
+  (let ((buffer (current-buffer))
+        (point (point)))
+    (setq vertico-posframe--buffer buffer)
+    (vertico-posframe--handle-minibuffer-window)
+    (vertico-posframe--show buffer point)))
 
 (defun vertico-posframe--show (buffer window-point)
   "`posframe-show' of vertico-posframe.
 
 BUFFER will be showed by `posframe-show'.  After `posframe-show'
 is called, window-point will be set to WINDOW-POINT."
-  (let ((posframe (apply #'posframe-show
-                         buffer
-                         :font vertico-posframe-font
-                         :poshandler vertico-posframe-poshandler
-                         :background-color (face-attribute 'vertico-posframe :background nil t)
-                         :foreground-color (face-attribute 'vertico-posframe :foreground nil t)
-                         :border-width vertico-posframe-border-width
-                         :border-color
-                         (let ((color (face-attribute 'vertico-posframe-border :background nil t)))
-                           (or (nth (- (minibuffer-depth) 1)
-                                    (list color "red" "yellow" "green" "blue" "purple"))
-                               color))
-                         :override-parameters vertico-posframe-parameters
-                         :refposhandler vertico-posframe-refposhandler
-                         :hidehandler #'vertico-posframe-hidehandler
-                         :lines-truncate t
-                         (funcall vertico-posframe-size-function))))
+  (let ((posframe
+         ;; Some posframe poshandlers need infos of last-window.
+         (with-selected-window (vertico-posframe-last-window)
+           (apply #'posframe-show
+                  buffer
+                  :font (buffer-local-value 'vertico-posframe-font buffer)
+                  ;; Variable settings in `vertico-multiform-commands' will
+                  ;; save to BUFFER as buffer-local variables, so we need to
+                  ;; get buffer local value from BUFFER, for example:
+                  ;;
+                  ;; (setq vertico-multiform-commands
+                  ;;       '((consult-line
+                  ;;          posframe
+                  ;;          (vertico-posframe-poshandler . posframe-poshandler-frame-top-center))
+                  ;;         (t buffer)))
+                  ;;
+                  :poshandler (buffer-local-value 'vertico-posframe-poshandler buffer)
+                  :background-color (face-attribute 'vertico-posframe :background nil t)
+                  :foreground-color (face-attribute 'vertico-posframe :foreground nil t)
+                  :border-width (buffer-local-value 'vertico-posframe-border-width buffer)
+                  :border-color (vertico-posframe--get-border-color)
+                  :override-parameters (buffer-local-value 'vertico-posframe-parameters buffer)
+                  :refposhandler (buffer-local-value 'vertico-posframe-refposhandler buffer)
+                  :hidehandler #'vertico-posframe-hidehandler
+                  :lines-truncate (buffer-local-value 'vertico-posframe-truncate-lines buffer)
+                  (funcall (buffer-local-value 'vertico-posframe-size-function buffer) buffer)))))
     ;; NOTE: `posframe-show' will force set window-point to 0, so we
     ;; need reset it again after `posframe-show'.
     (when (numberp window-point)
@@ -220,19 +266,46 @@ is called, window-point will be set to WINDOW-POINT."
       (setq-local cursor-type t)
       (setq-local cursor-in-non-selected-windows 'box))))
 
+(defun vertico-posframe--get-border-color ()
+  "Get color of vertico-posframe border."
+  (face-attribute
+   (let* ((n (minibuffer-depth))
+          (face (intern (format "vertico-posframe-border-%s" n)))
+          (face-fallback 'vertico-posframe-border-fallback))
+     (if (= n 1)
+         'vertico-posframe-border
+       (if (facep face)
+           face
+         face-fallback)))
+   :background))
+
 (defun vertico-posframe--show-minibuffer-p ()
   "Test show minibuffer or not."
   (cl-some
    (lambda (rule)
      (cond ((functionp rule)
             (funcall rule))
-           ((and rule (stringp rule))
+           ((and rule
+                 (stringp rule)
+                 (symbolp this-command))
             (string-match-p rule (symbol-name this-command)))
            ((symbolp rule)
             (symbol-value rule))
            (t nil)))
    vertico-posframe-show-minibuffer-rules))
 
+(defun vertico-posframe--handle-minibuffer-window ()
+  "Handle minibuffer window."
+  (let ((show-minibuffer-p (vertico-posframe--show-minibuffer-p))
+        (minibuffer-window (active-minibuffer-window)))
+    (setq-local max-mini-window-height 1)
+    (window-resize minibuffer-window
+                   (- (window-pixel-height minibuffer-window))
+                   nil nil 'pixelwise)
+    (set-window-vscroll minibuffer-window 100)
+    (when show-minibuffer-p
+      (set-window-vscroll minibuffer-window 0))))
+
 (defun vertico-posframe-last-window ()
   "Get the last actived window before active minibuffer."
   (let ((window (minibuffer-selected-window)))
@@ -241,20 +314,24 @@ is called, window-point will be set to WINDOW-POINT."
           (next-window))
         (selected-window))))
 
-(defun vertico-posframe--hide ()
-  "Hide vertico buffer."
-  (when (posframe-workable-p)
-    (posframe-hide vertico-posframe--buffer)))
+(defun vertico-posframe--minibuffer-exit-hook ()
+  "The function used by `minibuffer-exit-hook'."
+  ;; `vertico--resize-window' have set `max-mini-window-height' to
+  ;; 1.0, so I think setting it to 1.0 here is safe :-).
+  (setq-local max-mini-window-height 1.0)
+  (posframe-hide vertico-posframe--buffer))
 
 (defun vertico-posframe--setup ()
   "Setup minibuffer overlay, which pushes the minibuffer content down."
-  (add-hook 'minibuffer-exit-hook 'vertico-posframe--hide nil 'local))
+  (add-hook 'minibuffer-exit-hook #'vertico-posframe--minibuffer-exit-hook nil 'local))
 
 ;;;###autoload
 (defun vertico-posframe-cleanup ()
   "Remove frames and buffers used for vertico-posframe."
   (interactive)
-  (posframe-delete vertico-posframe--buffer))
+  (dolist (buffer (buffer-list))
+    (when (minibufferp buffer)
+      (posframe-delete-frame buffer))))
 
 ;;;###autoload
 (define-minor-mode vertico-posframe-mode
@@ -262,13 +339,17 @@ is called, window-point will be set to WINDOW-POINT."
   :global t
   (cond
    (vertico-posframe-mode
-    (advice-add #'vertico--display-candidates :after #'vertico-posframe--display)
-    (advice-add #'vertico--setup :after #'vertico-posframe--setup)
-    (advice-add #'vertico--resize-window :override #'ignore))
+    (if (not (posframe-workable-p))
+        (funcall (buffer-local-value 'vertico-posframe-fallback-mode (current-buffer)) 1)
+      (advice-add #'vertico--display-candidates :after #'vertico-posframe--display)
+      (advice-add #'vertico--setup :after #'vertico-posframe--setup)
+      (advice-add #'vertico--resize-window :override #'ignore)))
    (t
-    (advice-remove #'vertico--display-candidates #'vertico-posframe--display)
-    (advice-remove #'vertico--setup #'vertico-posframe--setup)
-    (advice-remove #'vertico--resize-window #'ignore))))
+    (if (not (posframe-workable-p))
+        (funcall (buffer-local-value 'vertico-posframe-fallback-mode (current-buffer)) -1)
+      (advice-remove #'vertico--display-candidates #'vertico-posframe--display)
+      (advice-remove #'vertico--setup #'vertico-posframe--setup)
+      (advice-remove #'vertico--resize-window #'ignore)))))
 
 (provide 'vertico-posframe)
 ;;; vertico-posframe.el ends here
